-- Amr Sabry and Zena Ariola
-- Jan 2005

-- Syntax and reductions of the l-mu-mubar-r-^ calculus

module L (
 C(..), V(..), E(..), T(..), W(..), 
 free,
 redv, rede, redc
) where

import List
import Monad

------------------------------------------------------------------------------
-- Syntax

data C = V `Bar` E | C `Csub` T | C `Cweak` W 
data V = Acc | Lam V | Mustack C | Muenv C | V `Vsub` T | V `Vweak` W 
data E = Tp | Stack | Env | V `Cons` E | Mub C | E `Esub` T | E `Eweak` W 
data T = Sacc V | Sstack E | Senv E 
data W = Wacc | Wstack | Wenv 

------------------------------------------------------------------------------
-- Free vars

binding :: T -> String
binding (Sacc _)   = "r"
binding (Sstack _) = "a"
binding (Senv _)   = "g"

class Free m where
  free :: m -> [String]

instance Free C where
  free (v `Bar` e) = free v `union` free e 
  free (Csub c t) = free t `union` delete (binding t) (free c)
  free (Cweak c w) = free c `union` free w

instance Free V where
  free Acc = ["r"]
  free (Lam v) = delete "r" (free v)
  free (Mustack c) = delete "a" (free c) 
  free (Muenv c) = delete "g" (free c)
  free (Vsub v t) = free t `union` delete (binding t) (free v)
  free (Vweak v w) = free v `union` free w

instance Free E where
  free Tp = []
  free Stack = ["a"] 
  free Env = ["g"]
  free (Cons v e) = free v `union` free e 
  free (Mub c) = delete "r" (free c)
  free (Esub e t) = free t `union` delete (binding t) (free e)
  free (Eweak e w) = free e `union` free w

instance Free T where
  free (Sacc v) = free v 
  free (Sstack e) = free e
  free (Senv e) = free e

instance Free W where
  free Wacc = ["r"]
  free Wstack = ["a"]
  free Wenv = ["g"]

------------------------------------------------------------------------------
-- Helper for reductions

has_ewacc (_ `Eweak` Wacc) = True
has_ewacc (e `Eweak` Wacc `Eweak` _) = True
has_ewacc _ = False

remove_ewacc (e `Eweak` Wacc) = e
remove_ewacc (e `Eweak` Wacc `Eweak` w) = e `Eweak` w
remove_ewacc e = e

-- 

has_ewstack (_ `Eweak` Wstack) = True
has_ewstack (e `Eweak` Wstack `Eweak` _) = True
has_ewstack _ = False

remove_ewstack (e `Eweak` Wstack) = e
remove_ewstack (e `Eweak` Wstack `Eweak` w) = e `Eweak` w
remove_ewstack e = e

-- 

has_ewenv (_ `Eweak` Wenv) = True
has_ewenv (e `Eweak` Wenv `Eweak` _) = True
has_ewenv _ = False

remove_ewenv (e `Eweak` Wenv) = e
remove_ewenv (e `Eweak` Wenv `Eweak` w) = e `Eweak` w
remove_ewenv e = e

--

has_vwstack (_ `Vweak` Wstack) = True
has_vwstack (v `Vweak` Wstack `Vweak` _) = True
has_vwstack _ = False

remove_vwstack (v `Vweak` Wstack) = v
remove_vwstack (v `Vweak` Wstack `Vweak` w) = v `Vweak` w
remove_vwstack v = v

-- 

has_vwenv (_ `Vweak` Wenv) = True
has_vwenv (v `Vweak` Wenv `Vweak` _) = True
has_vwenv _ = False

remove_vwenv (v `Vweak` Wenv) = v
remove_vwenv (v `Vweak` Wenv `Vweak` w) = v `Vweak` w
remove_vwenv v = v

------------------------------------------------------------------------
-- Reductions on terms which apply substitutions at the root

rv :: V -> Maybe V

rv (Acc `Vsub` (Sacc v')) = Just v'

rv (Lam v `Vsub` (Sstack e')) = 
    Just $ Lam (v `Vsub` (Sstack (e' `Eweak` Wacc)))
rv (Lam v `Vsub` (Senv e')) = 
    Just $ Lam (v `Vsub` (Senv (e' `Eweak` Wacc)))

rv (Mustack c `Vsub` (Sacc v')) = 
    Just $ Mustack (c `Csub` (Sacc (v' `Vweak` Wstack)))
rv (Muenv c `Vsub` (Sacc v')) = 
    Just $ Muenv (c `Csub` (Sacc (v' `Vweak` Wenv)))

rv (Mustack c `Vsub` (Senv e')) = 
    Just $ Mustack (c `Csub` (Senv (e' `Eweak` Wstack)))
rv (Muenv c `Vsub` (Sstack e')) = 
    Just $ Muenv (c `Csub` (Sstack (e' `Eweak` Wenv)))

rv (v `Vweak` Wacc `Vsub` (Sacc _)) = Just v
rv (v `Vweak` Wstack `Vsub` (Sstack _)) = Just v
rv (v `Vweak` Wenv `Vsub` (Senv _)) = Just v

rv (v `Vweak` Wacc `Vsub` (Sstack e)) | has_ewacc e = 
    Just $ (v `Vsub` (Sstack (remove_ewacc e))) `Vweak` Wacc
rv (v `Vweak` Wacc `Vsub` (Senv e)) | has_ewacc e = 
    Just $ (v `Vsub` (Senv (remove_ewacc e))) `Vweak` Wacc

rv (v `Vweak` Wstack `Vsub` (Senv e)) | has_ewstack e = 
    Just $ (v `Vsub` (Senv (remove_ewstack e))) `Vweak` Wstack
rv (v `Vweak` Wenv `Vsub` (Sstack e)) | has_ewenv e = 
    Just $ (v `Vsub` (Sstack (remove_ewenv e))) `Vweak` Wenv

rv (v `Vweak` Wstack `Vsub` (Sacc v')) | has_vwstack v' = 
    Just $ (v `Vsub` (Sacc (remove_vwstack v'))) `Vweak` Wstack
rv (v `Vweak` Wenv `Vsub` (Sacc v')) | has_vwenv v' = 
    Just $ (v `Vsub` (Sacc (remove_vwenv v'))) `Vweak` Wenv

rv _ = Nothing

-----
-- Allow reductions under two substitutions

redv v = (rv v) `mplus` (rv1 v) `mplus` (rv2 v) 

rv1 (v `Vsub` t) = do v' <- rv v; return $ v' `Vsub` t
rv1 _ = Nothing

rv2 (v `Vsub` t) = do v' <- rv1 v; return $ v' `Vsub` t
rv2 _ = Nothing
    
------------------------------------------------------------------------
-- Reductions on contexts which apply substitutions at the root

re :: E -> Maybe E 

re (Stack `Esub` (Sstack e')) = Just e'
re (Env `Esub` (Senv e')) = Just e'

re ((v `Cons` e) `Esub` t) = Just $ (v `Vsub` t) `Cons` (e `Esub` t)

re (Mub c `Esub` (Sstack e')) = 
    Just $ Mub (c `Csub` (Sstack (e' `Eweak` Wacc)))
re (Mub c `Esub` (Senv e')) = 
    Just $ Mub (c `Csub` (Senv (e' `Eweak` Wacc)))

re (e `Eweak` Wacc `Esub` (Sacc _)) = Just e
re (e `Eweak` Wstack `Esub` (Sstack _)) = Just e
re (e `Eweak` Wenv `Esub` (Senv _)) = Just e

re (e `Eweak` Wacc `Esub` (Sstack e')) | has_ewacc e' = 
    Just $ (e `Esub` (Sstack (remove_ewacc e'))) `Eweak` Wacc
re (e `Eweak` Wacc `Esub` (Senv e')) | has_ewacc e' = 
    Just $ (e `Esub` (Senv (remove_ewacc e'))) `Eweak` Wacc

re (e `Eweak` Wstack `Esub` (Senv e')) | has_ewstack e' = 
    Just $ (e `Esub` (Senv (remove_ewstack e'))) `Eweak` Wstack
re (e `Eweak` Wenv `Esub` (Sstack e')) | has_ewenv e' = 
    Just $ (e `Esub` (Sstack (remove_ewenv e'))) `Eweak` Wenv

re (e `Eweak` Wstack `Esub` (Sacc v)) | has_vwstack v = 
    Just $ (e `Esub` (Sacc (remove_vwstack v))) `Eweak` Wstack
re (e `Eweak` Wenv `Esub` (Sacc v)) | has_vwenv v = 
    Just $ (e `Esub` (Sacc (remove_vwenv v))) `Eweak` Wenv

re _ = Nothing

----- 
-- Allow reductions under two substitutions

rede e = (re e) `mplus` (re1 e) `mplus` (re2 e)

re1 (e `Esub` t) = do e' <- re e; return $ e' `Esub` t
re1 _ = Nothing

re2 (e `Esub` t) = do e' <- re1 e; return $ e' `Esub` t
re2 _ = Nothing
    
------------------------------------------------------------------------
-- Reductions on commands which apply substitutions at the root

rc :: C -> Maybe C 

rc ((v `Bar` e) `Csub` t) = 
    Just $ (v `Vsub` t) `Bar` (e `Esub` t)

rc (c `Cweak` Wacc `Csub` (Sacc _)) = Just c
rc (c `Cweak` Wstack `Csub` (Sstack _)) = Just c
rc (c `Cweak` Wenv `Csub` (Senv _)) = Just c

rc (c `Cweak` Wacc `Csub` (Sstack e)) | has_ewacc e = 
    Just $ (c `Csub` (Sstack (remove_ewacc e))) `Cweak` Wacc
rc (c `Cweak` Wacc `Csub` (Senv e)) | has_ewacc e = 
    Just $ (c `Csub` (Senv (remove_ewacc e))) `Cweak` Wacc

rc (c `Cweak` Wstack `Csub` (Senv e)) | has_ewstack e = 
    Just $ (c `Csub` (Senv (remove_ewstack e))) `Cweak` Wstack
rc (c `Cweak` Wenv `Csub` (Sstack e)) | has_ewenv e = 
    Just $ (c `Csub` (Sstack (remove_ewenv e))) `Cweak` Wenv

rc (c `Cweak` Wstack `Csub` (Sacc v)) | has_vwstack v = 
    Just $ (c `Csub` (Sacc (remove_vwstack v))) `Cweak` Wstack
rc (c `Cweak` Wenv `Csub` (Sacc v)) | has_vwenv v = 
    Just $ (c `Csub` (Sacc (remove_vwenv v))) `Cweak` Wenv

rc _ = Nothing

-----
-- Allow reductions under two substitutions

redc c = (rc c) `mplus` (rc1 c) `mplus` (rc2 c) 
 
rc1 (c `Csub` t) = do c' <- rc c; return $ c' `Csub` t
rc1 _ = Nothing

rc2 (c `Csub` t) = do c' <- rc1 c; return $ c' `Csub` t
rc2 _ = Nothing

------------------------------------------------------------------------------
-- Pretty-printing

instance Show C where
  show (v `Bar` e) = "<" ++ show v ++ " | " ++ show e ++ ">"
  show (Csub c t) = show c ++ show t
  show (Cweak c w) = show c ++ show w

instance Show V where
  show Acc = "r" 
  show (Lam v) = "(\\r. " ++ show v ++ ")"
  show (Mustack c) = "(mu s. " ++ show c ++ ")"
  show (Muenv c) = "(mu e. " ++ show c ++ ")"
  show (Vsub v t) = show v ++ show t
  show (Vweak v w) = show v ++ show w

instance Show E where
  show Tp = "tp" 
  show Stack = "s" 
  show Env = "e" 
  show (Cons v e) = "(" ++ show v ++ " . " ++ show e ++ ")"
  show (Mub c) = "(mub r. " ++ show c ++ ")"
  show (Esub e t) = show e ++ show t
  show (Eweak e w) = show e ++ show w

instance Show T where
  show (Sacc v) = "[r <- " ++ show v ++ "]"
  show (Sstack e) = "[s <- " ++ show e ++ "]"
  show (Senv e) = "[e <- " ++ show e ++ "]"

instance Show W where
  show Wacc = "^r"
  show Wstack = "^s"
  show Wenv = "^e"

------------------------------------------------------------------------------
