-- Amr Sabry and Zena Ariola
-- Jan 2005

-- Abstract machine

module NameAM (
  eval, embedI
) where

import L
import AI
import Debug.Trace

-------------------------------------------------------------------
-- Abstract machine

type State = (I, Env, Stack, Maybe Acc)
type Env = [Clos]
type Stack = [Clos]
type Acc = Clos
data Clos = Clos (I,Env) deriving Show

step :: State -> Maybe State
step (Exec, _, s, Just(Clos(c,e'))) = Just $ (c, e', s, Nothing)
step (Clear c, e, s, _) = Just $ (c, e, s, Nothing)
step (Closure c c', e, s, Nothing) = Just $ (c', e, s, Just(Clos(c,e)))
step (Pusharg c, e, s, Just(v)) = Just $ (c, e, v:s, Nothing)
step (Poparg c, e, v:s, Nothing) = Just $ (c, e, s, Just(v))
step (Extendenv c, e, s, Just(v)) = Just $ (c, v:e, s, Nothing)
step (Lookupenv c, v:e, s, Nothing) = Just $ (c, e, s, Just(v))
step _ = Nothing

eval :: I -> (Int,C)
eval is = loop 0 (is, [], [], Nothing)
  where loop i s =
            trace ("\n" ++ show s ++ "\n") $
            case step s of
              Nothing -> (i, embedstate s)
              Just s' -> loop (i+1) s'

-------------------------------------------------------------------
-- Embedding abstract machine states into lambda-mu-mu-r-^

embedstate :: State -> C
embedstate (c, e, s, Nothing) =
    (embedI c)
    `Csub`
    (Senv (embedenv e `Eweak` Wstack))
    `Csub`
    (Sstack (embedstack s))
embedstate (c, e, s, Just acc) =
    (embedI c)
    `Csub`
    (Senv (embedenv e `Eweak` Wstack `Eweak` Wacc))
    `Csub`
    (Sstack (embedstack s `Eweak` Wacc))
    `Csub`
    (Sacc (embedacc acc))

embedenv :: Env -> E
embedenv = foldr (Cons . embedacc) Tp

embedstack :: Stack -> E
embedstack = foldr (Cons . embedacc) Tp

embedacc :: Acc -> V
embedacc (Clos(c,e)) =
  Mustack (embedI c
           `Csub`
           (Senv (embedenv e `Eweak` Wstack)))

embedI :: I -> C
embedI Exec =
    (Acc `Vweak` Wstack `Vweak` Wenv)
    `Bar`
    (Stack `Eweak` Wacc `Eweak` Wenv)
embedI (Clear c) = embedI c `Cweak` Wacc
embedI (Closure c c') =
    ((Mustack (embedI c)) `Vweak` Wstack)
    `Bar`
    (Mub (embedI c'))
embedI (Pusharg c) =
    ((Mustack (embedI c)) `Vweak` Wstack `Vweak` Wacc)
    `Bar`
    ((Acc `Vweak` Wstack `Vweak` Wenv)
     `Cons`
     (Stack `Eweak` Wacc `Eweak` Wenv))
embedI (Poparg c) =
    ((Lam  (Mustack (embedI c))) `Vweak`  Wstack)
    `Bar`
    (Stack `Eweak` Wenv)
embedI (Extendenv c) =
    ((Muenv (embedI c)) `Vweak` Wacc `Vweak` Wenv)
    `Bar`
    ((Acc `Vweak` Wstack `Vweak` Wenv)
     `Cons`
     (Env `Eweak` Wstack `Eweak` Wacc))
embedI (Lookupenv c) =
    ((Lam (Muenv (embedI c))) `Vweak` Wenv)
    `Bar`
    (Env `Eweak` Wstack)

-------------------------------------------------------------------
